;	LPAGE.CMD:	LISP language MENU Page
;			for MicroEMACS 3.10 and above
;			(C)opyright 1989 by Gregory Wilcox
;

write-message "Loading LISP macros. . ."

set %tab 9
set %newline 13
set %space 32
set %dquote 34
set %lparen 40
set %rparen 41
set %indent "  " 

store-procedure skip-whitespace
	!while &or &or &equ $curchar %tab &equ $curchar %space &equ $curchar %newline
		forward-character
	!endwhile
	!endm 

store-procedure skip-ws-reverse
	!while &or &or &equ $curchar %tab &equ $curchar %space &equ $curchar %newline
		backward-character
	!endwhile
	!endm 

; if you try to do this using next-word, previous-word and end-of-word,
; you run into some knotty problems with characters like "*" and "-"
; which these functions ignore
store-procedure forward-atom
	set %paren-p FALSE
	set %white-p FALSE
	!while &not &or %paren-p %white-p
	set %paren-p &or &equal $curchar %lparen &equal $curchar %rparen
	set %white-p &or &equal $curchar %space &equal $curchar %tab
	forward-character
	!endwhile
	backward-character
	skip-whitespace
	!endm

store-procedure backward-atom
	set %paren-p FALSE
	set %white-p FALSE
	backward-character
	!while &not &or %paren-p %white-p
	set %paren-p &or &equal $curchar %lparen &equal $curchar %rparen
	set %white-p &or &equal $curchar %space &equal $curchar %tab
	backward-character
	!endwhile
	2 forward-character
	!endm

; forward-sexpr
27	store-macro
	skip-whitespace
	!if &equal $curchar %lparen
		goto-matching-fence	
		forward-character
	!else
		!if &equal $curchar %rparen
			forward-character
			write-message "[Leaving current S-expression]"
		!else
			; must be on an atom
			forward-atom
		!endif
	!endif
	!endm

; backward-sexpr
28	store-macro
	backward-character
	skip-ws-reverse
	!if &equal $curchar %rparen
		goto-matching-fence	
	!else
		!if &equal $curchar %lparen
			; no-op
			write-message "[Leaving current S-expression]"
		!else
			; must be on an atom
			backward-atom
		!endif
	!endif
	!endm

; mark-sexpr
29	store-macro
	!if &not &equal $curchar %lparen
		search-reverse "("
	!endif
	set-mark
	goto-matching-fence
	forward-character
	exchange-point-and-mark
	write-message "S-expression marked"
	!endm

; forward-function
30	store-macro
	!force search-forward "~n("
	!if &seq $status FALSE
		end-of-file
		write-message "Last function"
	!else
		backward-character
	!endif
	!endm

; backward-function
31	store-macro
	!force search-reverse "~n("
	!if &seq $status FALSE
		beginning-of-file
		write-message "First function"
	!else
	forward-character
	!endif
	!endm

; procedure pad
; replace leading whitespace on current line with %count copies of %indent
	store-procedure pad
	; trim leading whitespace
	!while &or &equ $curchar %space &equ $curchar %tab
		delete-next-character
	!endwhile		
	; now do the indenting
	!while &not &equ %count 0
		insert-string %indent
		set %count &sub %count 1
	!endwhile
	!endm

; indent-sexpr
32	store-macro
	write-message "Indenting S-expression. . ."
	forward-character
	execute-macro-26  ; backward-function
	; are we really here?
	!if &not &equal $curchar %lparen
		write-message "Can't find S-expression."
		!return
	!endif
	set %parens 0
*nest
	!if &equal $curchar %lparen
		; if sexpr is on one line, swallow it
		set %cline $curline
		set %ccol $curcol
		goto-matching-fence
		!if &not &equal %cline $curline
			set $curline %cline
			set $curcol %ccol
			set %parens &add %parens 1
		!endif
	!else
		!if &equal $curchar %rparen
			set %parens &sub %parens 1
		!endif
	!endif
	!if &equal $curchar %newline
		forward-character
		set %count %parens
		run pad
	!else
		forward-character
	!endif
	; skip comments (like this one)
	!if &equal $curchar %semicolon
		search-forward "~n"
		backward-character
	!endif
	!if &not &equ %parens 0
		!goto nest
	!endif
	write-message "Finished."
	!endm

; better-indent-sexpr
; strategy is to look only at parens, not letters
; this is not much faster than indent-sexpr
36	store-macro
	write-message "Indenting S-expression. . ."
	forward-character
	execute-macro-26  ; backward-function
	; are we really here?
	!if &not &equal $curchar %lparen
		write-message "Can't find S-expression."
		!return
	!endif
	set $discmd FALSE
	set %parens 0
*next
	run count-parens	
	set %count %parens
	run pad
	!if &not &equal %parens 0
		!goto next
	!endif
	set $discmd TRUE
	write-message "Done."
	!endm

; procedure count-parens
; counts the number of unmatched left parens on the current line
; result is in the user variable %parens
; leaves the cursor on the next line
	store-procedure count-parens
	beginning-of-line
	set-mark
	next-line
	copy-region
	set %line $kill
	set %length &length %line
*otog
	set %lpos &sindex %line "(" 
	set %rpos &sindex %line ")"
	!if &and &equal %lpos 0 &equal %rpos 0
		; write-message %parens
		!return
	!else
		!if &equal %lpos 0
			set %pos %rpos
			set %parens &sub %parens 1
		!else
			!if &equal %rpos 0
				set %pos %lpos
				set %parens &add %parens 1
			!else
				; take the first paren you come to
				!if &less %rpos %lpos
					set %pos %rpos
					set %parens &sub %parens 1
				!else
					set %pos %lpos
					set %parens &add %parens 1
				!endif
			!endif
		!endif
	!endif
	set %length &sub %length %pos
	set %line &right %line %length
	!goto otog
	!endm


; find-unbalanced-parentheses
33	store-macro
	; save cursor position
	set %cline $curline
	set %ccol $curcol
	write-message "Searching forward for mismatched parentheses. . ."
	beginning-of-file
*loop	
	!force search-forward "("
	!if &seq $status FALSE 
		!goto misup
	!endif
	backward-character
	!force goto-matching-fence
	!if &seq $status FALSE
		write-message "Too few close parentheses in this function."
		!return
	!endif
	!goto loop
*misup
	write-message "Searching backward for mismatched parentheses. . ."
	end-of-file
*loopr	
	!force search-reverse ")"
	!if &seq $status FALSE 
		!goto ok 
	!endif
	; forward-character not needed - asymmetry here
	!force goto-matching-fence
	!if &seq $status FALSE
		write-message "Too many close parentheses in this function."
		!return
	!endif
	!goto loopr
*ok	
	write-message "All parentheses appear balanced."
	; restore cursor position
	set $curline %cline
	set $curcol %ccol
	!endm

; toggle debug flag
34	store-macro
	set $debug &not $debug
	write-message $debug
	!endm

; load this buffer
35	store-macro
	set $debug FALSE
	execute-buffer $cbufname
	!endm

; run lisp
37	store-macro
	execute-program "xlisppc"
	!endm

; move-to-indent
; this is a simplistic version of the lispm tab key
38	store-macro 
	!if &equal $lwidth 0
		; indent to depth of previous line
		previous-line
		skip-whitespace
		set %ccol $curcol
		next-line
                !while &not &equal %ccol 0
                        insert-string " "
                        set %ccol &sub %ccol 1
                !endwhile
	!else
		skip-whitespace
	!endif
	!endm

; quick-arglist
; should check that the function name begins at the left paren
; (to avoid problems like finding foo when you wanted sfoo)
39	store-macro
	search-reverse "("
	next-word
	set-mark
	end-of-word
	copy-region
	set %function $kill
	set %argfile &find "d:\lisp\arglists.lsp"
	!if &seq %argfile ""
		write-message "Argument file not found"
		!return
	!endif
	set %cbuf $cbufname
	find-file %argfile
	beginning-of-file
	2 next-line	; skip past comments
*keepgoing
	!force search-forward %function
	!if &seq $status FALSE
		select-buffer %cbuf
		set %args &cat "Unknown function - " %function
		write-message %args
		!return
	!endif
	; check for end of function
	!if &not &equal $curchar %space
		!goto keepgoing
	!endif
	search-reverse "("
	set-mark
	end-of-line
	copy-region
	!force select-buffer %cbuf
	set %args &cat "Arglist = " $kill
	write-message %args
	; set $discmd TRUE
	!endm

; twiddle-sexprs
40	store-macro
	;kill the separator
	skip-whitespace
	set-mark
	backward-character
	!if &not &or &or &equal $curchar %space &equal $curchar %tab &equal $curchar %newline
		forward-character
		write-message "Separator should be <space>, <tab> or <newline>"
		!return
	!endif
	skip-ws-reverse
	forward-character
	kill-region
	set %sep $kill	
	execute-macro-22	; forward-sexpr
	kill-region
	execute-macro-23	; backward-sexpr
	yank
	insert-string %sep
	write-message "[S-expressions twiddled]"
	!endm

bind-to-key execute-macro-27 A-F	; forward-sexpr
bind-to-key execute-macro-28 A-B	; backward-sexpr
bind-to-key execute-macro-29 ^@		; mark-sexpr
bind-to-key execute-macro-30 A-E	; forward-function
bind-to-key execute-macro-31 A-A	; backward-function
bind-to-key execute-macro-32 A-I	; indent-sexpr
bind-to-key execute-macro-33 A-U	; find-unmatched-parentheses
bind-to-key execute-macro-34 A-D	; toggle debug flag
bind-to-key execute-macro-35 A-M	; load this buffer
bind-to-key execute-macro-36 A-J	; better-indent-sexpr
bind-to-key execute-macro-37 A-L	; run-lisp
bind-to-key execute-macro-38 S-^I	; move-to-indent
bind-to-key execute-macro-39 ^X^A	; quick-arglist 
bind-to-key execute-macro-40 A-T	; twiddle-sexprs

write-message "LISP macros are now loaded."


